PHIDATA_RUNTIME_ENV_VAR: str = "PHIDATA_RUNTIME"
PYTHONPATH_ENV_VAR: str = "PYTHONPATH"

SCRIPTS_DIR_ENV_VAR: str = "PHI_SCRIPTS_DIR"
STORAGE_DIR_ENV_VAR: str = "PHI_STORAGE_DIR"
META_DIR_ENV_VAR: str = "PHI_META_DIR"
PRODUCTS_DIR_ENV_VAR: str = "PHI_PRODUCTS_DIR"
NOTEBOOKS_DIR_ENV_VAR: str = "PHI_NOTEBOOKS_DIR"
WORKFLOWS_DIR_ENV_VAR: str = "PHI_WORKFLOWS_DIR"
WORKSPACE_ROOT_ENV_VAR: str = "PHI_WORKSPACE_ROOT"
WORKSPACE_MOUNT_ENV_VAR: str = "PHI_WORKSPACE_MOUNT"
WORKSPACE_CONFIG_DIR_ENV_VAR: str = "PHI_WORKSPACE_CONFIG_DIR"
WORKSPACE_CONFIG_FILE_ENV_VAR: str = "PHI_WORKSPACE_CONFIG_FILE"

AWS_REGION_ENV_VAR: str = "AWS_REGION"
AWS_DEFAULT_REGION_ENV_VAR: str = "AWS_DEFAULT_REGION"
AWS_PROFILE_ENV_VAR: str = "AWS_PROFILE"
AWS_CONFIG_FILE_ENV_VAR: str = "AWS_CONFIG_FILE"
AWS_SHARED_CREDENTIALS_FILE_ENV_VAR: str = "AWS_SHARED_CREDENTIALS_FILE"

INIT_AIRFLOW_ENV_VAR: str = "INIT_AIRFLOW"
AIRFLOW_ENV_ENV_VAR: str = "AIRFLOW_ENV"
AIRFLOW_EXECUTOR_ENV_VAR: str = "AIRFLOW__CORE__EXECUTOR"
AIRFLOW_DAGS_FOLDER_ENV_VAR: str = "AIRFLOW__CORE__DAGS_FOLDER"
AIRFLOW_DB_CONN_URL_ENV_VAR: str = "AIRFLOW__DATABASE__SQL_ALCHEMY_CONN"
